(* Time-stamp: <modified the 11/03/2019 (at 10:40) by Erwan Jahier> *)

(** Process programmer API *)
type varT = It | Ft | Bt | Et of int | St | Nt
type action = string (* just a label *)
type value = I of int | F of float | B of bool | E of int | S of string | N of int
type local_env = string -> value

type vars = (string * varT) list
type neighbor = {
  (*   n_id:  string; (* to hide? *) *)
  lenv:  local_env;
  n_vars: vars;
}

type enable_fun = neighbor list -> local_env -> action list
type step_fun   = neighbor list -> local_env -> action -> local_env

type int_tables = {
  vars: (string, vars)  Hashtbl.t;
  init_vars: (string, neighbor list -> local_env)  Hashtbl.t;
  enable: (string, enable_fun) Hashtbl.t;
  step: (string, step_fun) Hashtbl.t;
  actions: (string, action list) Hashtbl.t;
}



let tbls = {
  vars = Hashtbl.create 1;
  init_vars = Hashtbl.create 1;
  enable = Hashtbl.create 1;
  step = Hashtbl.create 1;
  actions = Hashtbl.create 1;
}

let verbose_level = ref 0

let vart_to_rif_string = function
  | It -> "int"
  | Ft -> "real"
  | Bt -> "bool"
  | St -> "string"
  | Et _i -> "int"
  | Nt -> "int"
    
let value_to_string = function
  | I i
  | E i
  | N i -> string_of_int i
  | F f -> string_of_float f
  | B true -> "t" 
  | B false -> "f" 
  | S str -> str
    
exception Unregistred of string * string

let print_table lbl tbl =
  let keys = Hashtbl.fold (fun k _ acc -> Printf.sprintf "%s,%s" k acc)  tbl "" in
  if !verbose_level > 0 then Printf.printf "Defined keys for %s: %s\n" lbl keys;
  flush stdout
    
type algo_id = string

let (reg_vars : algo_id  -> (string * varT) list -> unit) =
  fun algo_id x ->
  if !verbose_level > 0 then Printf.printf "Registering %s vars\n" algo_id;
  flush stdout;
  Hashtbl.replace tbls.vars  algo_id x
let (get_vars : string -> (string * varT) list) = fun algo_id ->  
  try Hashtbl.find tbls.vars  algo_id
  with Not_found ->
    print_table "vars" tbls.vars;
    raise (Unregistred ("variable", algo_id))

let (reg_init_vars : algo_id -> (neighbor list -> local_env) -> unit) = fun algo_id x ->
  if !verbose_level > 0 then Printf.printf "Registering %s init_vars\n" algo_id;
  flush stdout;
  Hashtbl.replace tbls.init_vars algo_id x
let (get_init_vars : algo_id -> (string * varT) list -> (neighbor list -> local_env)) =
  fun algo_id vars ->
    let default_env =
      (fun nl v ->
         match List.find_opt (fun (x,_t) -> x=v) vars with
           None -> failwith (v^" unknown var")
         | Some(_,Nt) ->
           assert (nl <> []);
           N (Random.int ((List.length nl)-1))
         | Some(_,It) -> I (Random.int 100000)
         | Some(_,Bt) -> B (Random.bool ())
         | Some(_,Ft) -> F (Random.float max_float)
         | Some(_,Et i) -> I (Random.int i)
         | Some(_,St) -> S "dummy"
                             
      )
    in
    try
      let user_env = Hashtbl.find tbls.init_vars algo_id in
      (fun nl v ->
         try user_env nl v
         with e ->
           if !verbose_level > 1 then
             Printf.eprintf "No init value for '%s' in user init function (%s).\n" v
               (Printexc.to_string e);
           default_env nl v)
    with Not_found ->
      if !verbose_level > 1 then
        Printf.eprintf "No user init function is provided.\n";
      default_env


let (reg_enable : algo_id -> enable_fun -> unit) = fun algo_id x -> 
        if !verbose_level > 0 then Printf.printf "Registering %s enable\n" algo_id;
        flush stdout;
        Hashtbl.replace tbls.enable  algo_id x
let (get_enable : algo_id -> enable_fun) = fun algo_id -> 
  try Hashtbl.find tbls.enable  algo_id
  with Not_found ->
    print_table "enable" tbls.enable;
    raise (Unregistred ("enable", algo_id))

let (reg_step : algo_id -> step_fun -> unit) = fun algo_id x -> 
  if !verbose_level > 0 then Printf.printf "Registering %s step\n" algo_id;
  flush stdout;
  Hashtbl.replace tbls.step  algo_id x
let (get_step : algo_id -> step_fun) = fun algo_id -> 
  try Hashtbl.find tbls.step  algo_id
  with Not_found ->
    print_table "step" tbls.step;
    raise (Unregistred ("step", algo_id))

let (reg_actions : algo_id -> action list -> unit) =
  fun algo_id x -> 
  if !verbose_level > 0 then Printf.printf "Registering %s actions\n" algo_id;
  flush stdout;
  Hashtbl.replace tbls.actions algo_id x
let (get_actions : algo_id -> action list) = fun algo_id -> 
  try Hashtbl.find tbls.actions  algo_id
  with Not_found ->
    print_table "actions" tbls.actions;
    raise (Unregistred ("actions", algo_id))
